! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2015 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                    List of ghost blocks                      !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Simple list of ghost blocks by utilizing the neighbor pointers of blocks.
!!
!! This list abuses the neighbor pointers of blocks to create a doubly
!! linked list. It only works for blocks whose neighbor pointers are
!! unused (ghost blocks, boundary ghost blocks).
!!
!! Two of the neighbor pointers are used:
!!>
!!    b%neigh(1,1)%l  ...  previous
!!    b%neigh(1,2)%l  ...  next
!!<
!!
!! Identification of ghost blocks:
!!>
!!    b%pos(1) == FD4_GHOST_BND   ... boundary ghost
!!    b%pos(1) == FD4_GHOST_CMM   ... communication ghost
!!<
!!
!!
!! @author Matthias Lieber
module fd4_ghostlist_mod

  use fd4_block_mod
  use fd4_globaldef_mod
  implicit none
  
  public
  
  integer(i_k), parameter :: FD4_GHOST_BND =  0  ! constant to identify boundary ghost blocks
  integer(i_k), parameter :: FD4_GHOST_CMM = -1  ! constant to identify communication ghost blocks
  
  type fd4_ghostlist
    type(fd4_block), pointer :: start => NULL()     !! first block of the ghostlist
    integer(i_k) :: length = 0                      !! length of the ghostlist
    logical :: touched = .false.                    !! indicator if ghostlist has been changed (set in add and remove)
  end type fd4_ghostlist

contains  
  


  !! Add block to the begining of the list.
  subroutine fd4_ghostlist_add(gl, b)
    implicit none
    type(fd4_ghostlist), intent(inout) :: gl    !! the ghostlist
    type(fd4_block), pointer           :: b     !! the block to add 
    
    if(associated(gl%start)) gl%start%neigh(1,1)%l => b
    b%neigh(1,1)%l => NULL()
    b%neigh(1,2)%l => gl%start
    gl%start => b
    gl%length = gl%length + 1
    gl%touched = .true.
    
  end subroutine fd4_ghostlist_add



  !! Remove specified block from the list.
  !! 
  !! Bad things will happen, if the block is not member of the list!
  subroutine fd4_ghostlist_remove(gl, b)
    implicit none
    type(fd4_ghostlist), intent(inout) :: gl    !! the ghostlist
    type(fd4_block), pointer           :: b     !! the block to remove    
    
    if(associated(b,gl%start)) gl%start => b%neigh(1,2)%l
    if(associated(b%neigh(1,1)%l)) b%neigh(1,1)%l%neigh(1,2)%l => b%neigh(1,2)%l
    if(associated(b%neigh(1,2)%l)) b%neigh(1,2)%l%neigh(1,1)%l => b%neigh(1,1)%l
    gl%length = gl%length - 1
    gl%touched = .true.
          
  end subroutine fd4_ghostlist_remove


  
  !! Get next block from the ghostlist.
  !!
  !!  Returns NULL() if $b is last block in the list. This routine simply does:
  !!v fd4_ghostlist_next => b%neigh(1,2)%l
  function fd4_ghostlist_next(b)
    implicit none
    type(fd4_block), pointer :: b                   !! the current block
    type(fd4_block), pointer :: fd4_ghostlist_next  !! return value: next block

    fd4_ghostlist_next => b%neigh(1,2)%l

  end function fd4_ghostlist_next
  
  
end module fd4_ghostlist_mod
