! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2015 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                           Meta data                          !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Management of distributed meta data for coupling.
!!
!! @author Matthias Lieber
module fd4_metadata_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod
  use fd4_mdsd_mod
  use fd4_block_mod
  use rbtree_fd4_mdsd_mod
  use rbtree_fd4_block_mod
  use rbtree_int_mod
  use stack3_mod
  use section_mod
  
  USE_MPI
  
  implicit none
  private

  !! shifting lengths for $fd4_metadata_c2i and $fd4_metadata_i2c
  integer, parameter :: m2i_shft3 = 20, m2i_shft2 = 10
  
  INCLUDE_MPIFH
  
  public :: fd4_metadata, fd4_mdsd_link
  public :: fd4_metadata_init, fd4_metadata_finalize
  public :: fd4_metadata_register, fd4_metadata_register_with_idx, fd4_metadata_unregister
  public :: fd4_metadata_full_allocate, fd4_metadata_clear
  public :: fd4_metadata_adapt, fd4_metadata_unadapt
  public :: fd4_metadata_get_section, fd4_metadata_get_subdomain, fd4_metadata_get_blbnd
  public :: fd4_metadata_blpos2mdidx, fd4_metadata_blpos2mdpos
  

  !! Type for the meta data context.
  type fd4_mdsd_link
    type(fd4_mdsd), pointer :: l => NULL()
  end type fd4_mdsd_link

  !! Type for the meta data context.
  type fd4_metadata
    integer(i_k) :: mdext(3) = 0         !! number of blocks per dimension of the domain
    integer(i_k) :: mdsdext(3) = 0       !! min extent of the meta data subdomains, i.e. number of blocks per dimension
    integer(i_k) :: mdsdnum(3) = 1       !! number of mdsd's per dimension
    integer(i_k) :: mdsdnum_tot = 0      !! max number of meta data subdomains
    type(rbtree_fd4_mdsd) :: mdsub       !! the meta data subdomains are managed in a tree
    type(fd4_mdsd_link), pointer :: mdbuf(:,:,:) => NULL() !! buffer of currently not required meta data subdomains
    integer(i_k) :: num_cpl(FD4_MD_MAX_PART) = 0 !! counts number of couple contexts per meta data partitioning
    logical :: single_mdsd = .false.     !! true if only one meta data subdomain is used (legacy mode)
    logical :: initialized = .false.
    logical :: adapt = .false.           !! true if meta data should adapt/optimize to block allocation
  end type fd4_metadata


contains

  include 'util/stack3.incf'


  !! Initialize meta data management.
  !!
  subroutine fd4_metadata_init(md, blnum, mdsdext, err)
    type(fd4_metadata), intent(inout) :: md   !! meta data context
    integer(i_k), intent(in)  :: blnum(3)     !! number of blocks per dimension of the domain
    integer(i_k), intent(in)  :: mdsdext(3)   !! min extent of the meta data subdomains, i.e. number of blocks per dimension
    integer(i_k), intent(out) :: err          !! error status: 0...ok
    
    err = -1
    FD4_L1_ASSERT(md%initialized, 'metadata already initialized')
    
    md%mdext(1:3) = blnum(1:3)
    md%mdsdext(1:3) = mdsdext(1:3)
    md%num_cpl = 0
    md%initialized = .true.
    ! get position of mdsd which contains the last block
    call fd4_metadata_blpos2mdpos(md, md%mdext, md%mdsdnum)
    ! to compute the global number of meta data subdomains
    md%mdsdnum_tot = PRODUCT(md%mdsdnum)
    md%single_mdsd = (md%mdsdnum_tot==1)
    
    err = 0
    
  end subroutine fd4_metadata_init


  !! Finalize meta data management.
  !!
  subroutine fd4_metadata_finalize(md)
    type(fd4_metadata), intent(inout) :: md   !! meta data context
  
    ! remove all meta data subdomains
    call rbtree_fd4_mdsd_clear (md%mdsub, fd4_mdsd_free)
    !if(fd4_wrnk==0) FD4_L2_INFO('removing all meta data subdomains')
    md%mdext(1:3) = 0
    md%mdsdext(1:3) = 0
    md%initialized = .false.

  end subroutine fd4_metadata_finalize


  !! Register a couple context with the meta data management.
  !!
  !! Returns $idx = -1 if max. number of couple context partitions is reached ($FD4_MD_MAX_PART).
  !!
  subroutine fd4_metadata_register(md, idx)
    type(fd4_metadata), intent(inout) :: md   !! meta data context
    integer(i_k), intent(out) :: idx          !! index of this couple context in the meta data
   
    idx = -1
    FD4_L1_ASSERT(.not.md%initialized,'meta data management not initialized')

    do idx=1,FD4_MD_MAX_PART
      if(md%num_cpl(idx)==0) then
        md%num_cpl(idx) = md%num_cpl(idx) + 1
        return
      end if
    end do

    idx = -1
    FD4_L1_ERROR('max. number of couple contexts is reached '//i2s(FD4_MD_MAX_PART))
    
  end subroutine fd4_metadata_register



  !! Register a couple context with the meta data management with a partition index
  !! of an already existing couple context.
  !! These couple contexts will share meta data.
  !!
  subroutine fd4_metadata_register_with_idx(md, idx)
    type(fd4_metadata), intent(inout) :: md   !! meta data context
    integer(i_k), intent(in) :: idx           !! index of this couple context in the meta data
    
    FD4_L1_ASSERT(.not.md%initialized,'meta data management not initialized')
    FD4_L1_ASSERT(md%num_cpl(idx)==0,'no partition with this idx registered yet')
    FD4_L3_ASSERT(idx<1,'idx < 1')
    FD4_L3_ASSERT(idx>FD4_MD_MAX_PART,'idx > FD4_MD_MAX_PART')
    
    md%num_cpl(idx) = md%num_cpl(idx) + 1

  end subroutine fd4_metadata_register_with_idx
  


  !! Unregister a couple context with the meta data management.
  !!
  !! Removes memory associated with this couple context.
  !! When the last couple context unregisters, all meta data subdomains are removed.
  !!
  subroutine fd4_metadata_unregister(md, idx)
    type(fd4_metadata), intent(inout) :: md
    integer(i_k), intent(in) :: idx

    type(fd4_mdsd), pointer :: mdsd
    type(rbtree_fd4_mdsd_iter) :: mditer

    FD4_L3_ASSERT(idx<1,'md idx < 1')
    FD4_L3_ASSERT(idx>FD4_MD_MAX_PART,'md idx > max')
    
    md%num_cpl(idx) = md%num_cpl(idx) - 1
    FD4_L3_ASSERT(md%num_cpl(idx)<0,'md%num_cpl(idx) < 0')

    if( md%num_cpl(idx) == 0 ) then
      ! this is the last couple context with this partition

      ! loop over mdsds and remove entries of this couple context only
      mdsd => rbtree_fd4_mdsd_iter_first(mditer, md%mdsub)
      do while(associated(mdsd))
        call stack3_delete(mdsd%desc_rank(idx))
        call stack3_delete(mdsd%desc_abnd(idx))
        mdsd => rbtree_fd4_mdsd_iter_next(mditer)
      end do

    end if
    
    
  end subroutine fd4_metadata_unregister


  !! Removes all memory associated with the meta data management.
  !!
  subroutine fd4_metadata_clear(md)
    type(fd4_metadata), intent(inout) :: md

    call rbtree_fd4_mdsd_clear(md%mdsub, fd4_mdsd_free)
    if(associated(md%mdbuf)) deallocate(md%mdbuf)
    md%initialized = .false.
    
  end subroutine fd4_metadata_clear


  !! Allocates the all meta data subdomains.
  !!
  subroutine fd4_metadata_full_allocate(md)
    type(fd4_metadata), intent(inout) :: md   !! meta data context

    integer(i_k) :: mdpos(3), mdidx, i, j, k
    type(fd4_mdsd), pointer :: mdsd
    
    FD4_L3_ASSERT(md%mdsub%length>md%mdsdnum_tot,'too much mdsds')
    if(md%mdsub%length>=md%mdsdnum_tot) return
    
    ! loop over all mdsd positions
    do k=1,md%mdsdnum(3)
      mdpos(3)=k
      do j=1,md%mdsdnum(2)
        mdpos(2)=j
        do i=1,md%mdsdnum(1)
          mdpos(1)=i

          mdidx = fd4_metadata_c2if(mdpos)
          mdsd => rbtree_fd4_mdsd_get(md%mdsub, mdidx)
          if(.not. associated(mdsd)) then
            ! insert new meta data subdomain at mdpos
            allocate(mdsd)
            call rbtree_fd4_mdsd_insert(md%mdsub, mdidx, mdsd)
          end if

        end do
      end do
    end do
    
  end subroutine fd4_metadata_full_allocate
  

  !! Get the meta data subdomain which contains the block at $blpos.
  !!
  !! Returns NULL() if the mdsd is not allocated.
  !!
  subroutine fd4_metadata_get_subdomain(md, blpos, mdsd)
    type(fd4_metadata), intent(in) :: md   !! meta data context
    integer(i_k), intent(in) :: blpos(3)   !! position of the block
    type(fd4_mdsd), pointer :: mdsd        !! the meta data subdomain

    integer(i_k) :: mdpos(3)

    call fd4_metadata_blpos2mdpos(md, blpos, mdpos)
    mdsd => rbtree_fd4_mdsd_get(md%mdsub, fd4_metadata_c2if(mdpos))

  end subroutine fd4_metadata_get_subdomain


  !! Get a section (list of rectangles) describing the union of bounds of
  !! the allocated meta data subdomains.
  !!
  subroutine fd4_metadata_get_section(md, sect)
    type(fd4_metadata), intent(in) :: md      !! meta data context
    type(section), intent(out) :: sect        !! position of the block
    
    type(fd4_mdsd), pointer :: mdsd
    type(rbtree_fd4_mdsd_iter) :: mditer
    integer(i_k) :: mdpos(3), blbnd(3,2)

    FD4_L3_ASSERT(md%mdsub%length>md%mdsdnum_tot,'too much mdsds')

    call section_reset(sect)
    
    if(md%mdsub%length>=md%mdsdnum_tot) then

      ! all mdsd's allocated
      blbnd(1:3,1) = 1
      blbnd(1:3,2) = md%mdext
      call section_add(sect, blbnd)

    else
    
      mdsd => rbtree_fd4_mdsd_iter_first(mditer, md%mdsub)
      do while(associated(mdsd))
        call fd4_metadata_i2pos(mditer%cur%key, mdpos)
        call fd4_metadata_mdpos2blbnd(md, mdpos, blbnd)
        call section_add(sect, blbnd)
        mdsd => rbtree_fd4_mdsd_iter_next(mditer)
      end do

    end if

  end subroutine fd4_metadata_get_section
  


  !! Move required meta data subdomains from the buffer to the acutal mdsd
  !! storage and vice versa after repartitioning.
  !!
  subroutine fd4_metadata_adapt(md, blocks)
    type(fd4_metadata),     intent(inout) :: md        !! meta data context
    type(rbtree_fd4_block), intent(inout) :: blocks    !! blocks of the domain
    
    integer(i_k) :: mdidx, mdbnd(3,2), mdpos(3), i, j, k
    type(rbtree_fd4_mdsd) :: mdtmp
    type(fd4_mdsd), pointer :: mdsd
    type(rbtree_fd4_mdsd_iter) :: mditer
    type(fd4_block), pointer :: b
    type(rbtree_fd4_block_iter) :: biter

    if( (.not. md%initialized) .or. md%single_mdsd) return
    
    if(.not. associated(md%mdbuf)) then
      allocate(md%mdbuf(md%mdsdnum(1),md%mdsdnum(2),md%mdsdnum(3)))
    end if

    !! 1: move all mdsds to the buffer
    
    ! copy all mdsd pointers from mdsub to mdbuf
    mdsd => rbtree_fd4_mdsd_iter_first(mditer, md%mdsub)
    do while(associated(mdsd))
      call fd4_metadata_i2ijk(mditer%cur%key, i, j, k)
      md%mdbuf(i,j,k)%l => mdsd
      mdsd => rbtree_fd4_mdsd_iter_next(mditer)
    end do
    ! remove all mdsd from mdsub
    call rbtree_fd4_mdsd_detach(md%mdsub)
    
    !! 2: move the required mdsds from the buffer to the mdsd storage
    
    ! loop over all local blocks
    b => rbtree_fd4_block_iter_first(biter, blocks)
    do while(associated(b))
      ! this is the position of the mdsd which covers the current block
      call fd4_metadata_blpos2mdpos(md, b%pos, mdpos)
      ! check if this mdsd is still contained in mdbuf
      if(associated(md%mdbuf(mdpos(1),mdpos(2),mdpos(3))%l)) then
        ! get index of this mdsd and insert it into mdsub
        mdidx = fd4_metadata_c2if(mdpos)
        call rbtree_fd4_mdsd_insert(md%mdsub, mdidx, md%mdbuf(mdpos(1),mdpos(2),mdpos(3))%l)
        ! remove mdsd from mdbuf
        md%mdbuf(mdpos(1),mdpos(2),mdpos(3))%l => NULL()
        !write(fd4_out,'(A,A,I12,A,A,I6)') FD4_LINEHD,'insert: ',mdidx,' at ',p2s(b%pos),md%mdsub%length
      end if
      b => rbtree_fd4_block_iter_next(biter)
    end do
    
    !! 3: fulfill criteria for necessary mdsds:
    !!    add those which lie ahead in space (also diagonal) of current mdsds
    
    ! loop over all current mdsds
    mdsd => rbtree_fd4_mdsd_iter_first(mditer, md%mdsub)
    do while(associated(mdsd))
      ! loop over neighbor mdsds ahead of current mdsd
      call fd4_metadata_i2pos(mditer%cur%key, mdbnd(1:3,2))
      mdbnd(1:3,1) = MAX(1, mdbnd(1:3,2) - 1)
      do k=mdbnd(3,1),mdbnd(3,2)
        do j=mdbnd(2,1),mdbnd(2,2)
          do i=mdbnd(1,1),mdbnd(1,2)
            ! check if this neighbor mdsd is still contained in mdbuf
            if(associated(md%mdbuf(i,j,k)%l)) then
              ! get index of this mdsd and insert it into mdtmp
              mdidx = fd4_metadata_ijk2if(i,j,k)
              call rbtree_fd4_mdsd_insert(mdtmp, mdidx, md%mdbuf(i,j,k)%l)
              ! remove mdsd from mdbuf
              md%mdbuf(i,j,k)%l => NULL()
              !write(fd4_out,'(A,A,I12,I6)') FD4_LINEHD,'insert: ',mdidx,md%mdsub%length
            end if
          end do
        end do
      end do
      mdsd => rbtree_fd4_mdsd_iter_next(mditer)
    end do
    ! copy all mdsd pointers from mdtmp to mdsub
    mdsd => rbtree_fd4_mdsd_iter_first(mditer, mdtmp)
    do while(associated(mdsd))
      call rbtree_fd4_mdsd_insert(md%mdsub, mditer%cur%key, mdsd)
      mdsd => rbtree_fd4_mdsd_iter_next(mditer)
    end do
    ! remove all mdsd from mdtmp
    call rbtree_fd4_mdsd_detach(mdtmp)
  
    FD4_L4_INFO('adapted meta data '//i2s(md%mdsub%length))

  end subroutine fd4_metadata_adapt


  !! Move alle meta data subdomains from the buffer to the acutal mdsd storage.
  !! (Undo the adaption/optimization from $fd4_metadata_adapt)
  !!
  subroutine fd4_metadata_unadapt(md)
    type(fd4_metadata),     intent(inout) :: md        !! meta data context
    
    integer(i_k) :: mdidx, i, j, k
    
    if( (.not. md%initialized) .or. md%single_mdsd .or. .not. associated(md%mdbuf) ) return

    ! copy all mdsd pointers from mdbuf to mdsub
    do k=1,md%mdsdnum(3)
      do j=1,md%mdsdnum(2)
        do i=1,md%mdsdnum(1)
          ! check if this neighbor mdsd is still contained in mdbuf
          if(associated(md%mdbuf(i,j,k)%l)) then
            ! get index of this mdsd and insert it into mdtmp
            mdidx = fd4_metadata_ijk2if(i,j,k)
            call rbtree_fd4_mdsd_insert(md%mdsub, mdidx, md%mdbuf(i,j,k)%l)
            ! remove mdsd from mdbuf
            md%mdbuf(i,j,k)%l => NULL()
            !write(fd4_out,'(A,A,I12,I6)') FD4_LINEHD,'unadapt: ',mdidx,md%mdsub%length
          end if
        end do
      end do
    end do
  
    FD4_L4_INFO('unadapted meta data')

  end subroutine fd4_metadata_unadapt



  !! Get the min and max block position of a meta data subdomain identified by $mdidx.
  !!
  pure subroutine fd4_metadata_get_blbnd(md, mdidx, blbnd)
    type(fd4_metadata), intent(in) :: md     !! meta data context
    integer(i_k), intent(in)  :: mdidx       !! index of the meta data subdomain
    integer(i_k), intent(out) :: blbnd(3,2)  !! block bounds of the mdsd
    
    integer(i_k) :: mdpos(3)

    call fd4_metadata_i2pos(mdidx, mdpos)
    call fd4_metadata_mdpos2blbnd(md, mdpos, blbnd)

  end subroutine fd4_metadata_get_blbnd


  !! Get the index for a meta data subdomain in which the block at $blpos resides.
  !!
  pure integer(i_k) function fd4_metadata_blpos2mdidx(md, blpos)
    implicit none
    type(fd4_metadata), intent(in) :: md     !! meta data context
    integer(i_k), intent(in)  :: blpos(3)    !! position of the block
    integer(i_k) :: mdpos(3)
    call fd4_metadata_blpos2mdpos(md, blpos, mdpos)
    fd4_metadata_blpos2mdidx = fd4_metadata_c2if(mdpos)
  end function fd4_metadata_blpos2mdidx





  !! Get the position of a meta data subdomain in which the block at $blpos resides.
  !!
  pure subroutine fd4_metadata_blpos2mdpos(md, blpos, mdpos)
    type(fd4_metadata), intent(in) :: md     !! meta data context
    integer(i_k), intent(in)  :: blpos(3)    !! position of the block
    integer(i_k), intent(out) :: mdpos(3)    !! position of the meta data subdomain
    mdpos(1:3) = (blpos(1:3)-1) / md%mdsdext(1:3) + 1
  end subroutine fd4_metadata_blpos2mdpos


  !! Get the min and max block position of a meta data subdomain.
  !!
  pure subroutine fd4_metadata_mdpos2blbnd(md, mdpos, blbnd)
    type(fd4_metadata), intent(in) :: md     !! meta data context
    integer(i_k), intent(in)  :: mdpos(3)    !! position of the meta data subdomain
    integer(i_k), intent(out) :: blbnd(3,2)  !! min and max block position
    blbnd(1:3,2) = mdpos(1:3) * md%mdsdext(1:3)
    blbnd(1:3,1) = blbnd(1:3,2) - md%mdsdext(1:3) + 1
    blbnd(1:3,2) = MIN(blbnd(1:3,2), md%mdext)
  end subroutine fd4_metadata_mdpos2blbnd


  !! Mapping from coordinate to index for meta data subdomains
  !!
  !! Same method as in $fd4_domain_c2i ($fd4_domain.incf).
  !!
  elemental pure integer(i_k) function fd4_metadata_ijk2if(i, j, k)
    implicit none
    integer(i_k), intent(in) :: i,j,k
    fd4_metadata_ijk2if = ishft(k,m2i_shft3) + ishft(j,m2i_shft2) + i
  end function fd4_metadata_ijk2if

  !! Mapping from coordinate to index for meta data subdomains
  !!
  !! Same method as in $fd4_domain_c2i ($fd4_domain.incf).
  !!
  pure integer(i_k) function fd4_metadata_c2if(pos)
    implicit none
    integer(i_k), intent(in) :: pos(3)
    fd4_metadata_c2if = ishft(pos(3),m2i_shft3) + ishft(pos(2),m2i_shft2) + pos(1)
  end function fd4_metadata_c2if

  !! Mapping from index to coordinate for meta data subdomains
  !!
  !! Same method as in $fd4_domain_i2c ($fd4_domain.incf).
  !! 
  elemental pure subroutine fd4_metadata_i2ijk(idx, i, j, k)
    implicit none
    integer(i_k), intent(in)  :: idx
    integer(i_k), intent(out) :: i, j, k
    i = iand( idx , 2**m2i_shft2-1 )
    j = iand( ishft( idx , -m2i_shft2 ) , 2**(m2i_shft3-m2i_shft2)-1 )
    k = ishft( idx , -m2i_shft3 )
  end subroutine fd4_metadata_i2ijk

  !! Mapping from index to coordinate for meta data subdomains
  !!
  !! Same method as in $fd4_domain_i2c ($fd4_domain.incf).
  !! 
  pure subroutine fd4_metadata_i2pos(idx, pos)
    implicit none
    integer(i_k), intent(in)  :: idx
    integer(i_k), intent(out) :: pos(3)
    pos(1) = iand( idx , 2**m2i_shft2-1 )
    pos(2) = iand( ishft( idx , -m2i_shft2 ) , 2**(m2i_shft3-m2i_shft2)-1 )
    pos(3) = ishft( idx , -m2i_shft3 )
  end subroutine fd4_metadata_i2pos

end module fd4_metadata_mod
