! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2015 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

! DO NOT CHANGE THESE ROUTINES WITHOUT THE SAME CHANGES IN stack.incf !

!! A simple stack of 3D coordinates (3 integers) which grows ad infinutum if required.
!! Data is stored in a single contiguous array.
!!
!!> type(stack3) :: st
!!  call stack3_create(st,10)
!!  stack3_push(st,(/4,1,2/))
!!  stack3_push(st,(/7,0,5/))
!!  print *,'first element on stack is:',st%a(:,1)
!!< call stack3_delete(st)
!!
!!
!!  Clear the stack, i.e. set topmost element = bottom.
!!  The size of the storage array is left unchanged.
!!v st%pos = 0
!!
!!
!!  Get element from stack.
!!  The index must be in the interval 1 <= $i <= size(st%a).
!!v st%a(:,i)
!!
!!  Add
!!v use stack3_mod
!!  to your module/subroutine and
!!v include 'stack3.incf'
!!  after the contains statement to enable inlining of the stack3 routines.
!!
!! @author Matthias Lieber
module stack3_mod
  
  use kinds_mod
  implicit none
  
  private
  
  public :: stack3

  type stack3
    integer(i_k), pointer :: a(:,:) => null()  !! the actual data is stored in $a
    integer(i_k) :: pos = 0                    !! position of topmost element in the stack
  end type stack3

contains

  !! Create a stack3 or reallocate at given size.
  !!
  !! If the stack3 is already allocated, the function reallocates
  !! the stack3 only in the case, that the current stack size is
  !! less then $inisize. Otherwise the stack3 is only cleared,
  !! i.e. $st%pos is set to 0.
  subroutine stack3_create(st, inisize)
    type(stack3), intent(inout) :: st         !! the stack3
    integer(i_k), intent(in) :: inisize      !! initial size of array
    integer(i_k) :: is 
    is = MAX(inisize,1)
    if( associated(st%a) ) then
      if( ubound(st%a,2) < is ) then
        deallocate( st%a )
        allocate( st%a(3,is) )
      end if
    else
      allocate( st%a(3,is) )
    end if    
    st%pos = 0
  end subroutine stack3_create

  !! Delete the stack3 and all of its memory.
  subroutine stack3_delete(st)
    type(stack3), intent(inout) :: st    !! the stack3
    if(associated(st%a)) deallocate( st%a )
    st%a => null()
    st%pos = 0
  end subroutine stack3_delete


  !! Reallocate the stack3 array to exactly fit the required size.
  subroutine stack3_trim(st)
    type(stack3), intent(inout) :: st    !! the stack3
    integer(i_k), pointer       :: a(:,:)
    if(.not. associated(st%a)) return
    if(st%pos<ubound(st%a,2)) then
      ! allocate smaller array
      allocate( a(3,MAX(1,st%pos)) )
      a(1:3,1:st%pos) = st%a(1:3,1:st%pos)
      deallocate(st%a)
      st%a => a
    end if
  end subroutine stack3_trim  


  !! Push a new element to the top of the stack3.
  !! The storage array will be enlarged by a factor of two if required.
  subroutine stack3_push(st, v)
    type(stack3), intent(inout) :: st    !! the stack3
    integer(i_k), intent(in)    :: v(3)      !! value to push to stack3
    integer(i_k), pointer       :: a(:,:)
    if(st%pos>=ubound(st%a,2)) then
      ! allocate bigger array
      allocate( a(3,2*st%pos) )
      a(1:3,1:st%pos) = st%a(1:3,1:st%pos)
      deallocate(st%a)
      st%a => a
      !print *,'allocated array of size',2*st%pos
    end if
    st%pos = st%pos + 1
    st%a( 1:3, st%pos ) = v
  end subroutine stack3_push  

  !! Move (shallow copy) the data from $st_src to $st_dest.
  !! All data contained in $st_dest will be removed.
  !! $st_src will be empty on return.
  subroutine stack3_move(st_src, st_dest)
    type(stack3), intent(inout) :: st_src     !! the stack3
    type(stack3), intent(inout) :: st_dest    !! the stack3
    call stack3_delete(st_dest)
    ! shallow copy
    st_dest%a   => st_src%a
    st_dest%pos =  st_src%pos    
    ! empty st_src
    st_src%a   => null()
    st_src%pos = 0
  end subroutine stack3_move

  
  !! Remove duplicates from $st, this is damn costly
  subroutine stack3_remove_duplicates(st)
    type(stack3), intent(inout) :: st     !! the stack3
    integer :: i, j
    !! start of fancy O(N^2) algorithm
    i=1
    do while (i<st%pos)
      j=st%pos
      ! why starting at the end? because when swapping a hit with the last
      ! element, we are sure that the last element is no duplicate.
      do while (j>i)
        if( ALL( st%a(:,i) == st%a(:,j) ) ) then
          !! hit! swap with last element
          if(j<st%pos) st%a(:,j) = st%a(:,st%pos)
          st%pos = st%pos - 1
        end if
        j=j-1
      end do
      i=i+1
    end do    
  end subroutine stack3_remove_duplicates


end module stack3_mod
