! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2015 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

! DO NOT CHANGE THESE ROUTINES WITHOUT THE SAME CHANGES IN stack.incf !

!! A simple stack of 3D coordinates (3 integers) which grows ad infinutum if required.
!! Data is stored in a single contiguous array.
!!
!! In contrast to the $stack3, the $stack3s stores the current top of the stack $pos
!! in the same contiguous memory as the actual data, which makes it easily possible
!! to transfer the $stack3s via message passing.
!!
!!> type(stack3s) :: st
!!  call stack3s_create(st,10)
!!  stack3s_push(st,(/4,1,2/))
!!  stack3s_push(st,(/7,0,5/))
!!  print *,'first element on stack is:',st%a(:,1)
!!< call stack3s_delete(st)
!!
!!
!!  Clear the stack, i.e. set topmost element = bottom.
!!  The size of the storage array is left unchanged.
!!v st%pos = 0
!!
!!
!!  Get element from stack.
!!  The index must be in the interval 1 <= $i <= size(st%a).
!!v st%a(:,i)
!!
!!  Add
!!v use stack3s_mod
!!  to your module/subroutine and
!!v include 'stack3s.incf'
!!  after the contains statement to enable inlining of the stack3s routines.
!!
!! @see stack3_mod
!! @author Matthias Lieber
module stack3s_mod
  
  use kinds_mod
  implicit none
  
  private
  
  public :: stack3s

  type stack3s
    integer(i_k), pointer :: a(:,:) => null()  !! the data is stored in $a
    integer(i_k), pointer :: pos => null()     !! position of topmost element in the stack
    integer(i_k), pointer :: m(:,:) => null()  !! here we allocate the actual memory for both $a and $pos
  end type stack3s

contains

  !! Create a stack3s or reallocate at given size.
  !!
  !! If the stack3s is already allocated, the function reallocates
  !! the stack3s only in the case, that the current stack size is
  !! less then $inisize. Otherwise the stack3s is only cleared,
  !! i.e. $st%pos is set to 0.
  subroutine stack3s_create(st, inisize)
    type(stack3s), intent(inout) :: st         !! the stack3s
    integer(i_k), intent(in) :: inisize      !! initial size of array
    integer(i_k) :: is 
    is = MAX(inisize,1)
    if( associated(st%a) ) then
      if( ubound(st%a,2) < is ) then
        deallocate( st%m )
      end if
    end if
    if( .not. associated(st%a) ) then
      allocate( st%m(3,0:is) )
      st%a => st%m(1:3,1:is)
      st%pos => st%m(1,0)
    end if    
    st%pos = 0
  end subroutine stack3s_create

  !! Delete the stack3s and all of its memory.
  subroutine stack3s_delete(st)
    type(stack3s), intent(inout) :: st    !! the stack3s
    if(associated(st%m)) deallocate( st%m )
    st%m => null()
    st%a => null()
    st%pos => null()
  end subroutine stack3s_delete


  !! Reallocate the stack3s array to exactly fit the required size.
  subroutine stack3s_trim(st)
    type(stack3s), intent(inout) :: st    !! the stack3s
    integer(i_k), pointer       :: m(:,:)
    if(.not. associated(st%m)) return
    if(st%pos<ubound(st%a,2)) then
      ! allocate smaller array
      allocate( m(3,0:MAX(1,st%pos)) )
      m(1:3,1:st%pos) = st%m(1:3,1:st%pos)
      m(1,0) = st%pos
      st%a => m(1:3,1:st%pos)
      st%pos => m(1,0)
      deallocate(st%m)
      st%m => m
    end if
  end subroutine stack3s_trim  


  !! Push a new element to the top of the stack3s.
  !! The storage array will be enlarged by a factor of two if required.
  subroutine stack3s_push(st, v)
    type(stack3s), intent(inout) :: st    !! the stack3s
    integer(i_k), intent(in)    :: v(3)      !! value to push to stack3s
    integer(i_k), pointer       :: m(:,:)
    if(st%pos>=ubound(st%a,2)) then
      ! allocate bigger array
      allocate( m(3,0:2*st%pos) )
      m(1:3,1:st%pos) = st%m(1:3,1:st%pos)
      m(1,0) = st%pos
      st%a => m(1:3,1:2*st%pos)
      st%pos => m(1,0)
      deallocate(st%m)
      st%m => m
      !print *,'allocated array of size',2*st%pos
    end if
    st%pos = st%pos + 1
    st%a( 1:3, st%pos ) = v
  end subroutine stack3s_push  

end module stack3s_mod
